local type = type
local ipairs = ipairs
local Talented = Talented
local GameTooltip = GameTooltip
local IsAltKeyDown = IsAltKeyDown
local TOOLTIP_TALENT_RANK = TOOLTIP_TALENT_RANK
local TOOLTIP_TALENT_TIER_POINTS = TOOLTIP_TALENT_TIER_POINTS
local TOOLTIP_TALENT_PREREQ = TOOLTIP_TALENT_PREREQ
local TOOLTIP_TALENT_NEXT_RANK = TOOLTIP_TALENT_NEXT_RANK
local GREEN_FONT_COLOR = GREEN_FONT_COLOR
local NORMAL_FONT_COLOR = NORMAL_FONT_COLOR
local HIGHLIGHT_FONT_COLOR = HIGHLIGHT_FONT_COLOR
local RED_FONT_COLOR = RED_FONT_COLOR
local TOOLTIP_TALENT_LEARN = TOOLTIP_TALENT_LEARN

local L = LibStub("AceLocale-3.0"):GetLocale("Talented")

local function addline(line, color, split)
	GameTooltip:AddLine(line, color.r, color.g, color.b, split)
end

local function add2line(left, right)
	GameTooltip:AddDoubleLine(left, right, GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b, GREEN_FONT_COLOR.r, GREEN_FONT_COLOR.g, GREEN_FONT_COLOR.b)
end

local function addtipline(tip, color)
	color = color or NORMAL_FONT_COLOR
	tip = tip or ""
	if type(tip) == "string" then
		addline(tip, color, true)
	else
		for _, i in ipairs(tip) do
			if i.right then
				GameTooltip:AddDoubleLine(i.left, i.right, color.r, color.g, color.b, color.r, color.g, color.b)
			else
				addline(i.left, color, true)
			end
		end
	end
end

local spellTooltip
local function CreateSpellTooltip()
	local tt = CreateFrame"GameTooltip"
	local lefts = {}
	for i = 1, 20 do
		local left, right = tt:CreateFontString(), tt:CreateFontString()
		left:SetFontObject(GameFontNormal)
		right:SetFontObject(GameFontNormal)
		tt:AddFontStrings(left, right)
		lefts[i] = left
	end
	tt.lefts = lefts
	function tt:SetSpell(spell)
		if self.spell ~= spell then
			self:SetOwner(TalentedFrame)
			self:ClearLines()
			self:SetHyperlink("spell:"..spell)
			self.spell = spell
		end
	end
	Talented.spellNameCache = setmetatable({}, {
		__index = function (self, key)
			tt:SetSpell(key)
			local value = tt.lefts[1]:GetText()
			self[key] = value
			return value
		end,
	})
	Talented.spellDescCache = setmetatable({}, {
		__index = function (self, key)
			tt:SetSpell(key)
			local value = tt.lefts[tt:NumLines()]:GetText()
			self[key] = value
			return value
		end,
	})
	CreateSpellTooltip = nil
	return tt
end

function Talented:GetSpellName(class, tab, index)
	if not spellTooltip then
		spellTooltip = CreateSpellTooltip()
	end
	local spell = self:UncompressSpellData(class)[tab][index][1]
	return self.spellNameCache[spell]
end

function Talented:GetSpellDesc(class, tab, index, rank)
	if not spellTooltip then
		spellTooltip = CreateSpellTooltip()
	end
	local spell = self:UncompressSpellData(class)[tab][index][rank]
	return self.spellDescCache[spell]
end

local lastTooltipFrame
function Talented:SetTooltipInfo(frame, tab, index)
	if not tab or not index then return end
	lastTooltipFrame = frame
	if not GameTooltip:IsOwned(frame) then
		GameTooltip:SetOwner(frame, "ANCHOR_RIGHT")
	end

	local tree = self.talents[self.viewclass][tab]
	local info = tree.talents[index].info
	GameTooltip:ClearLines()
	if tiptype ~= "table" and self.template == self.current then
		GameTooltip:SetTalent(tab, index)
		GameTooltip:Show()
		return
	end
	local tier = (info.row - 1) * 5
	self:UnpackTemplate(self.template)
	local rank = self.template[tab][index]
	addline(self:GetSpellName(self.viewclass, tab, index), HIGHLIGHT_FONT_COLOR)
	addline(TOOLTIP_TALENT_RANK:format(value, info.ranks), HIGHLIGHT_FONT_COLOR)
	if info.prereqs then
		for _, req in ipairs(info.prereqs) do
			local o = tree.talents[req.source].info
			if self.template[tab][req.source] < o.ranks then
				addline(TOOLTIP_TALENT_PREREQ:format(o.ranks, self:GetSpellName(self.viewclass, tab, req.source)), RED_FONT_COLOR)
			end
		end
	end
	if tier >= 1 and self:GetTalentTabCount(self.template, tab) < tier then
		addline(TOOLTIP_TALENT_TIER_POINTS:format(tier, tree.info.name), RED_FONT_COLOR)
	end
	if IsAltKeyDown() then
		for i = 1, info.ranks do
			addtipline(self:GetSpellDesc(self.viewclass, tab, index, i), i == rank and HIGHLIGHT_FONT_COLOR or NORMAL_FONT_COLOR)
		end
	else
		if rank > 0 then
			addtipline(self:GetSpellDesc(self.viewclass, tab, index, rank))
		end
		if rank < info.ranks then
			addline(TOOLTIP_TALENT_NEXT_RANK, HIGHLIGHT_FONT_COLOR)
			addtipline(self:GetSpellDesc(self.viewclass, tab, index, rank + 1))
		end
	end
	local s = self:GetTalentState(self.template, tab, index)
	if self.mode == "edit" then
		if self.template == self.current then
			if s == "available" or s == "empty" then
				addline(TOOLTIP_TALENT_LEARN, GREEN_FONT_COLOR)
			end
		elseif s == "full" then
			addline(L["Right-click to unlearn"], GREEN_FONT_COLOR)
		elseif s == "available" then
			add2line(TOOLTIP_TALENT_LEARN, L["Right-click to unlearn"])
		elseif s == "empty" then
			addline(TOOLTIP_TALENT_LEARN, GREEN_FONT_COLOR)
		end
	end
	GameTooltip:Show()
end

function Talented:HideTooltipInfo()
	GameTooltip:Hide()
	lastTooltipFrame = nil
end

function Talented:UpdateTooltip()
	if lastTooltipFrame and GameTooltip:IsShown() and  GameTooltip:IsOwned(lastTooltipFrame) then
		self:SetTooltipInfo(lastTooltipFrame, lastTooltipFrame.tab, lastTooltipFrame.id)
	end
end

function Talented:MODIFIER_STATE_CHANGED(_, mod)
	if mod:sub(-3) == "ALT" then
		self:UpdateTooltip()
	end
end
